"""
Extension for the `TaskCog`.
Loading this file via `commands.Bot.load_extension` will add `TaskCog` to the bot.
"""

import json
import logging
import datetime
from os import getenv
from time import process_time

import aiohttp
from discord import TextChannel
from discord import app_commands
from discord.ext import commands, tasks
from discord.errors import Forbidden
from sqlalchemy import insert, select, and_
from feedparser import parse

from feed import Source, Article, RSSFeed, Subscription, SubscriptionChannel, SubChannel
from db import (
    DatabaseManager,
    FeedChannelModel,
    RssSourceModel,
    SentArticleModel
)
from utils import get_unparsed_feed
from api import API

log = logging.getLogger(__name__)

TASK_INTERVAL_MINUTES = getenv("TASK_INTERVAL_MINUTES")

# task trigger times : must be of type list
times = [
    datetime.time(hour, minute, tzinfo=datetime.timezone.utc)
    for hour in range(24)
    for minute in range(0, 60, int(TASK_INTERVAL_MINUTES))
]

log.debug("Task will trigger every %s minutes", TASK_INTERVAL_MINUTES)


class TaskCog(commands.Cog):
    """
    Tasks cog.
    """

    def __init__(self, bot):
        super().__init__()
        self.bot = bot
        self.time = None

    @commands.Cog.listener()
    async def on_ready(self):
        """Instructions to execute when the cog is ready."""

        if not self.bot.developing:
            self.rss_task.start()

        log.info("%s cog is ready", self.__class__.__name__)

    @commands.Cog.listener(name="cog_unload")
    async def on_unload(self):
        """Instructions to execute before the cog is unloaded."""

        self.rss_task.cancel()

    @app_commands.command(name="debug-trigger-task")
    async def debug_trigger_task(self, inter):
        await inter.response.defer()
        await self.rss_task()
        await inter.followup.send("done")

    @tasks.loop(time=times)
    async def rss_task(self):
        """Automated task responsible for processing rss feeds."""

        log.info("Running subscription task")
        time = process_time()

        guild_ids = [guild.id for guild in self.bot.guilds]
        data = []

        async with aiohttp.ClientSession() as session:
            api = API(self.bot.api_token, session)
            page = 0

            while True:
                page += 1
                page_data = await self.get_subscriptions(api, guild_ids, page)

                if not page_data:
                    break

                data.extend(page_data)
                log.debug("extending data by '%s' items", len(page_data))

            log.debug("finished api data collection, browsed %s pages for %s subscriptions", page, len(data))

            subscriptions = Subscription.from_list(data)
            for sub in subscriptions:
                await self.process_subscription(api, session, sub)

            log.info("Finished subscription task, time elapsed: %s", process_time() - time)

    async def get_subscriptions(self, api, guild_ids: list[int], page: int):

        log.debug("attempting to get subscriptions for page: %s", page)

        try:
            return (await api.get_subscriptions(server__in=guild_ids, page=page))[0]
        except aiohttp.ClientResponseError as error:
            if error.status == 404:
                log.debug(error)
                return []

            log.error(error)

    async def process_subscription(self, api, session, sub: Subscription):
        log.debug("processing subscription '%s' '%s' for '%s'", sub.id, sub.name, sub.guild_id)

        if not sub.active:
            log.debug("skipping sub because it's active flag is 'False'")
            return

        channels = [self.bot.get_channel(subchannel.channel_id) for subchannel in await sub.get_channels(api)]
        if not channels:
            log.warning("No channels to send this to")
            return

        unparsed_content = await get_unparsed_feed(sub.url, session)
        parsed_content = parse(unparsed_content)
        source = Source.from_parsed(parsed_content)
        articles = source.get_latest_articles(3)

        if not articles:
            log.debug("No articles found")

        for article in articles:
            await self.process_article(session, channels, article)

    async def process_article(self, session, channels: list[SubChannel], article: Article):
        embed = await article.to_embed(session)

        log.debug("attempting to send embed to %s channel(s)", len(channels))

        for channel in channels:
            await channel.send(embed=embed)


    # async def batch_process_subs(self, data: list):

    #     log.debug("batch process subs, count '%s'", len(data))
    #     subscriptions = Subscription.from_list(data)

    #     for sub in subscriptions:
    #         log.info(sub.name)


    # async def process_feed(self, feed: FeedChannelModel, database: DatabaseManager):
    #     """Process the passed feed. Will also call process for each article found in the feed.

    #     Parameters
    #     ----------
    #     feed : FeedChannelModel
    #         Database model for the feed.
    #     database : DatabaseManager
    #         Database connection handler, must be open.
    #     """

    #     log.debug("Processing feed: %s", feed.id)

    #     channel = self.bot.get_channel(feed.discord_channel_id)

    #     # TODO: integrate the `validate_feed` code into here, also do on list command and show errors.

    #     async with aiohttp.ClientSession() as session:

    #         unparsed_content = await get_unparsed_feed(feed.rss_source.rss_url)
    #         parsed_feed = parse(unparsed_content)
    #         source = Source.from_parsed(parsed_feed)
    #         articles = source.get_latest_articles(5)

    #         if not articles:
    #             log.info("No articles to process for %s in ", feed.rss_source.nick, feed.discord_server_id)
    #             return

    #         for article in articles:
    #             await self.process_article(feed.id, article, channel, database, session)

    # async def process_article(
    #     self, feed_id: int, article: Article, channel: TextChannel, database: DatabaseManager,
    #     session: aiohttp.ClientSession
    # ):
    #     """Process the passed article. Will send the embed to a channel if all is valid.

    #     Parameters
    #     ----------
    #     feed_id : int
    #         The feed model ID, used to log the sent article.
    #     article : Article
    #         Database model for the article.
    #     channel : TextChannel
    #         Where the article will be sent to.
    #     database : DatabaseManager
    #         Database connection handler, must be open.
    #     """

    #     log.debug("Processing article: %s", article.url)

    #     query = select(SentArticleModel).where(and_(
    #         SentArticleModel.article_url == article.url,
    #         SentArticleModel.discord_channel_id == channel.id,
    #     ))
    #     result = await database.session.execute(query)

    #     if result.scalars().all():
    #         log.debug("Article already processed: %s", article.url)
    #         return

    #     embed = await article.to_embed(session)
    #     try:
    #         await channel.send(embed=embed)
    #     except Forbidden as error:  # TODO: find some way of informing the user about this error.
    #         log.error("Can't send article to channel: %s · %s · %s", channel.name, channel.id, error)
    #         return

    #     query = insert(SentArticleModel).values(
    #         article_url = article.url,
    #         discord_channel_id = channel.id,
    #         discord_server_id = channel.guild.id,
    #         discord_message_id = -1,
    #         feed_channel_id = feed_id
    #     )
    #     await database.session.execute(query)

    #     log.debug("new Article processed: %s", article.url)


async def setup(bot):
    """
    Setup function for this extension.
    Adds `TaskCog` to the bot.
    """

    cog = TaskCog(bot)
    await bot.add_cog(cog)
    log.info("Added %s cog", cog.__class__.__name__)
